#!/usr/bin/perl -w
# This program is a client for the proprietary Bootcast protocol used by
# Kreatel set top boxes.
#
# Copyright 2004 by Marco d'Itri <md@linux.it>
# This software is distributed under the terms of the GNU GPL. If we meet some
# day, and you think this stuff is worth it, you can buy me a beer in return.

use strict;
use Socket;

# XXX FIXME please implement command line parsing with Getopt::Long
my $verbose = 1;
my $port = 11111;
my $group = '239.113.254.1';
my $filename = 'output';

# type flags packet-number total-packets filename
use constant BOOTCAST_PACKET => 'n n N N A[32] a*';

# from <netinet/in.h>, verified on Linux x86
use constant IPPROTO_IP => 0;
use constant IP_ADD_MEMBERSHIP => 35;

my $sock = open_multicast($group, $port);
print "Multicast socket opened.\n";

open(OUT, ">$filename") or die $!;

my $waiting = 1;
my $expected = 1;

while (1) {
	my $buf;
	my $addr = recv($sock, $buf, 2 ** 16, 0);
	die "recv: $!" if not defined $addr;

	my ($type, $flags, $seq, $total, $name, $data)
		= unpack(BOOTCAST_PACKET, $buf);

	if ($seq != $expected) {
		if ($waiting) {
			print "Still " . ($total - $seq) . " packets left to skip.\n"
				if $verbose and ($seq % 100) == 0;
		} else {
			print "WARNING: packet $seq has been lost (received $expected)\n";
			$waiting = 1;
		}
		next;
	}
	$waiting = 0;

	print "Received $total packets.\n" if $verbose and ($seq % 100) == 0;

	print OUT $data;
	$expected++;

	if ($seq == $total) {
		my ($srcport, $srcip) = sockaddr_in($addr);
		my $srcaddr = inet_ntoa($srcip);
		print "Received file $name from $srcaddr:$srcport.\n"
			. "($total packets, type $type, flags $flags)\n";
		last;
	}
}

close $sock;
close OUT or die "close: $!";
exit 0;

sub open_multicast {
	my ($group, $port) = @_;

	#if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)

	socket(SOCK, PF_INET, SOCK_DGRAM, 0) or die $!;

	#saddr.sin_family = AF_INET;
	#saddr.sin_addr.s_addr = inet_addr(argv[0]);
	#saddr.sin_port = htons(atoi(argv[1]));
	#if (bind(sock, (struct sockaddr *) &saddr, sizeof(saddr)) < 0)

	bind(SOCK, sockaddr_in($port, inet_aton($group))) or die $!;

	#mr.imr_multiaddr = saddr.sin_addr;
	#mr.imr_interface.s_addr = htonl(INADDR_ANY);
	#if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mr, sizeof(mr)) < 0)

	#struct in_addr imr_multiaddr;       /* IP multicast address of group */
	#struct in_addr imr_interface;       /* local IP address of interface */
	my $ip_mreq = inet_aton($group) . inet_aton('0.0.0.0');
	setsockopt(SOCK, IPPROTO_IP, IP_ADD_MEMBERSHIP, $ip_mreq) or die $!;

	return \*SOCK;
}

